/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    TECMULTITONEDETECT.C

    A simple example console application demonstrating the use the following
    devices:
    
      OTX_LDEVICE_TONE_DTMF_DETECTOR, 
      OTX_LDEVICE_TONE_DUAL_DETECTOR
      OTX_LDEVICE_TONE_SILENCE_DETECTOR
      OTX_LDEVICE_TONE_SINGLE_DETECTOR, 

   DTMF / DualTone / Silence / Single tones are detected.
    
   The applications works with the following board combinations:

      Thor-2-ExpressCard
      Thor-4-ExpressCard

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 E1 Lines. It creates various tone devices on the DSP, 
    which are used to detect tones. 
    
    The total number of logical devices supported in this demo:

      DTMF Detector: 60, DualTone Detector: 60, 
      Silence Detector: 60, SingleTone Detector: 60

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"


#define DEMO_DEFAULT_LOG_FILE "TecMultiToneDetect.log"

#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_NUMBER_OF_DEVS   60         //
#define DEMO_NO_OF_LIS        4          // Thor-2-ExpressCard, Thor-4-ExpressCard is tested in this demo                    

typedef struct DemoEnvT {
  char            *pszInputStr;
  OTX_BOOL        bExitOnError;
  OTX_INT32       nBoardType;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  int             boardNo;
  OtxDataFormatE  law;
  OtxT1E1LiModeE  eT1E1Mode;                // T1 or E1 mode
  OtxDspSpmE      eProgramModule;           // Standard Program Module to be loaded in the DSP (from the Lib)
  int             coDsp;
  int             coLi;
  int             nIteration;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);

void *DemoDtmfDetector(void *lpParameter);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_UINT32  g_nDspHighwaySpeed;  // DSP Highway speed in kpbs

OTX_BOOL  g_sqDev[DEMO_NUMBER_OF_DEVS] = {OTX_FALSE};
char g_sqDialDigit[DEMO_NUMBER_OF_DEVS][20];


struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hDualE1T1[1]; // Thor-4-ExpressCard has a QuadFalc
  OTX_HANDLE hLi[DEMO_NO_OF_LIS];
  OTX_HANDLE hDsp[1];
} g_hPhysDevs;


struct {
  OTX_INT32 nDev;

  OTX_INT32  nTs;
  OTX_INT32  nLi;
  OTX_INT32  nDsp;

  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_HANDLE hHandle;   // logical device
  OTX_HANDLE hLiDspPipe;

  OTX_BOOL bActive;
  OTX_BOOL bVerbose;

} g_sqDtmfDetector[DEMO_NUMBER_OF_DEVS];


struct {
  OTX_INT32 nDev;

  OTX_INT32  nTs;
  OTX_INT32  nLi;
  OTX_INT32  nDsp;

  OTX_HANDLE hHandle;   // logical device

  OTX_BOOL bActive;
  OTX_BOOL bVerbose;

  clock_t startTime, stopTime;

} g_sqDualToneDetector[DEMO_NUMBER_OF_DEVS];


struct {
  OTX_INT32 nDev;

  OTX_INT32  nTs;
  OTX_INT32  nLi;
  OTX_INT32  nDsp;

  OTX_HANDLE hHandle;   // logical device

  OTX_BOOL bActive;
  OTX_BOOL bVerbose;

  clock_t startTime, stopTime;

} g_sqSilenceDetector[DEMO_NUMBER_OF_DEVS];


struct {
  OTX_INT32 nDev;

  OTX_INT32  nTs;
  OTX_INT32  nLi;
  OTX_INT32  nDsp;

  OTX_HANDLE hHandle;   // logical device

  OTX_BOOL bActive;
  OTX_BOOL bVerbose;

  clock_t startTime, stopTime;

} g_sqSingleToneDetector[DEMO_NUMBER_OF_DEVS];


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;



/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func
    Initalize the global data for the DTMF / DualTone Dialer and Detector devices

*******************************************************************************/
void DemoInitializeGlobalData(
  void
)
{
  OTX_INT32 i;

  // Connect the first 30 devices to Li0 (TS1-15, and TS17-31, skip TS16)
  // Connect the remaining 30 devices to Li1 (TS1-15, and TS17-31, skip TS16)

  // DTMF
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    g_sqDtmfDetector[i].nDev = i;  
    g_sqDtmfDetector[i].nLi  = i/30;

    if ((i % 30) < 15) { // TS01 - TS15
      g_sqDtmfDetector[i].nTs  = (i % 30) + 1;  // not using TS0
    } else {      // TS17 - TS31
      g_sqDtmfDetector[i].nTs  = (i % 30) + 2;  // not using TS16
    }

    g_sqDtmfDetector[i].nDsp = 0;                          
    g_sqDtmfDetector[i].bActive = OTX_FALSE;
  }

  // print out
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    DemoLogf("DTMF Detector: RX[%.2d] Li=%.2d LiTs=%.2d Dsp=%.2d \n", 
      i,
      g_sqDtmfDetector[i].nLi,       
      g_sqDtmfDetector[i].nTs,       
      g_sqDtmfDetector[i].nDsp      
    );
  }

  
  // DualTone
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    g_sqDualToneDetector[i].nDev = i;
    g_sqDualToneDetector[i].nLi  = i/30;

    if ((i % 30) < 15) { // TS01 - TS15
      g_sqDualToneDetector[i].nTs  = (i % 30) + 1;  // not using TS0
    } else { // TS17 - TS31
      g_sqDualToneDetector[i].nTs  = (i % 30) + 2;  // not using TS16
    }

    g_sqDualToneDetector[i].nDsp = 0;                          
    g_sqDualToneDetector[i].bActive = OTX_FALSE;
  }
  
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    DemoLogf("Dual Tone Detector: RX[%.2d] Li=%.2d LiTs=%.2d Dsp=%.2d \n", 
      i,
      g_sqDualToneDetector[i].nLi,       
      g_sqDualToneDetector[i].nTs,       
      g_sqDualToneDetector[i].nDsp      
    );
  }


  // Silence Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    g_sqSilenceDetector[i].nDev = i;
    g_sqSilenceDetector[i].nLi  = i/30;

    if ((i % 30) < 15) { // TS01 - TS15
      g_sqSilenceDetector[i].nTs  = (i % 30) + 1;  // not using TS0 
    } else { // TS17 - TS31
      g_sqSilenceDetector[i].nTs  = (i % 30) + 2;  // not using TS16
    }

    g_sqSilenceDetector[i].nDsp = 0;                          
    g_sqSilenceDetector[i].bActive = OTX_FALSE;
  }

  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    DemoLogf("Silence Detector: RX[%.2d] Li=%.2d LiTs=%.2d Dsp=%.2d \n", 
      i,
      g_sqSilenceDetector[i].nLi,       
      g_sqSilenceDetector[i].nTs,       
      g_sqSilenceDetector[i].nDsp      
    );
  }


  // SingleTone Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    g_sqSingleToneDetector[i].nDev = i;
    g_sqSingleToneDetector[i].nLi  = i/30;

    if ((i % 30) < 15) { // TS01 - TS15
      g_sqSingleToneDetector[i].nTs  = (i % 30) + 1;  // not using TS0 
    } else { // TS17 - TS31
      g_sqSingleToneDetector[i].nTs  = (i % 30) + 2;  // not using TS16
    }

    g_sqSingleToneDetector[i].nDsp = 0;                          
    g_sqSingleToneDetector[i].bActive = OTX_FALSE;
  }

  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    DemoLogf("Single Tone Detector: RX[%.2d] Li=%.2d LiTs=%.2d Dsp=%.2d \n", 
      i,
      g_sqSingleToneDetector[i].nLi,       
      g_sqSingleToneDetector[i].nTs,       
      g_sqSingleToneDetector[i].nDsp      
    );
  }


}


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // create a notification event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);           

  // create event queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // enable event queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nDsp;
  OTX_INT32  nDualE1T1; 
  OTX_INT32  nQuadE1T1; 

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    DemoLogf("Opening Thor-2-ExpressCard board\n");
    break;
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    DemoLogf("Opening Thor-4-ExpressCard board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE)g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Get handles to T1/E1 line interface devices (#0 and #1)
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    {
      // On the Thor-2-ExpressCard board the two E1T1 devices are hosted by a dual E1T1 device. Open those that device.
      nDualE1T1 = 0; 
      DemoLogf("Open OTX_DEVICE_DUAL_T1E1 #%d\n", nDualE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DUAL_T1E1, nDualE1T1, nDualE1T1, g_hEventQueue, &g_hPhysDevs.hDualE1T1[nDualE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nDualE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nDualE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hDualE1T1[nDualE1T1], OTX_DEVICE_LI_T1E1, i%2, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hDualE1T1[nDualE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    // Thor-4-ExpressCard by default is set to receive on the four lines 
    // (monitor mode) with all transmitters disabled (tristate). This 
    // demo however needs the card to support the transmit data so the 
    // card will use 2 receivers and the 2 transmitters.
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController0, 
                                         OTX_ATTR_BRD_MONITOR_ONLY, 
                                         OTX_FALSE); 
    DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Configure Thor-4-ExpressCard for data transmit and receive\n");

    // we'll source the Falc from the DSP device, and source the DSP from the FALC
    //  FALC <-- DSP
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController0, 
                                           OTX_ATTR_BRD_FALC_SOURCED, 
                                           OTX_BRD_FALC_SOURCED_DSP); 
    DemoCheckRc(g_hPhysDevs.hThorController0, result);
    //  DSP <-- FALC
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController0, 
                                           OTX_ATTR_BRD_DSP_SOURCED, 
                                           OTX_BRD_DSP_SOURCED_FALC); 
    DemoCheckRc(g_hPhysDevs.hThorController0, result);

    DemoLogf("  Set FPGA highway path DSP->Falc and Falc->DSP\n");

    // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. 
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hDualE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    
    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hDualE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hDualE1T1[nQuadE1T1], result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  // Open DSP    
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  } 

  // Get DSP Highway Speed
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[i], OTX_ATTR_DSP_HIGHWAY_SPEED, &g_nDspHighwaySpeed); 
    DemoLogf("DSP#%d highway speed is %d kbps\n", i, g_nDspHighwaySpeed);
  }

  // Load DSP Program from OtxDrv.lib
  for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM2_INTMEM:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2_INTMEM\n", i);
      break;
    default:
      DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[nDsp], g_demoEnv.eProgramModule); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  }
    
}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // configure transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // set termination for the T1/E1 front end
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}



/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nDualE1T1; 
  OTX_INT32  nQuadE1T1; 

  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // If board type is Thor-2-ExpressCard then enable Dual E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
    // Enable Dual E1T1 device
    nDualE1T1 = 0; 
    DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nDualE1T1);
    result = OtxDrvEnable(g_hPhysDevs.hDualE1T1[nDualE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDualE1T1[nDualE1T1], result);
  }
  else // If board type is Thor-4-ExpressCard then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hDualE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDualE1T1[nQuadE1T1], result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // DTMF Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
      
    // create a notification event and event queue for the DTMF Detector devices
    result = OtxDrvCreateEvent(&g_sqDtmfDetector[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_sqDtmfDetector[i].hNotificationEvent, &g_sqDtmfDetector[i].hEventQueue); DemoCheckRc(g_sqDtmfDetector[i].hEventQueue, result);
    result = OtxDrvEnable(g_sqDtmfDetector[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDetector[i].hEventQueue, result);

    // create DTMF Detector logical devices
    result = OtxDrvCreateLogicalDevice( 
      g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], 
      OTX_LDEVICE_TONE_DTMF_DETECTOR, 
      g_sqDtmfDetector[i].hEventQueue, 
      i,                  
      &g_sqDtmfDetector[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], result);

  }


  // Dual Tone Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {

    // create DualTone Detector logical devices
    result = OtxDrvCreateLogicalDevice( 
      g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], 
      OTX_LDEVICE_TONE_DUAL_DETECTOR, 
      g_hEventQueue, 
      i,                  
      &g_sqDualToneDetector[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], result);

  }


  // Silence Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
      
    // create Silence Detector logical devices
    result = OtxDrvCreateLogicalDevice( 
      g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], 
      OTX_LDEVICE_TONE_SILENCE_DETECTOR, 
      g_hEventQueue,  
      i,                  
      &g_sqSilenceDetector[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], result);
  }


  // SingleTone Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
      
    // create SingleTone Detector logical devices
    result = OtxDrvCreateLogicalDevice( 
      g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], 
      OTX_LDEVICE_TONE_SINGLE_DETECTOR, 
      g_hEventQueue,  
      i,                  
      &g_sqSingleToneDetector[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i%g_demoEnv.coDsp], result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 nDev;
  OTX_INT32 nTsLi;
  OTX_INT32 nTsDsp;
  OTX_INT32 nLi;
  OTX_INT32 nDsp;
  OTX_UINT32 coTsPerDspHwy;
  OTX_RESULT result;
  OTX_UINT8 sqLiTimeSlotMask[32];
  OTX_UINT8 sqDspTimeSlotMask[128];

  coTsPerDspHwy = g_nDspHighwaySpeed / 64;


  // DTMF devices connection
  for (nDev = 0; nDev < DEMO_NUMBER_OF_DEVS; nDev++) {
    result = OtxDrvCreatePipe(64, &g_sqDtmfDetector[nDev].hLiDspPipe); DemoCheckRc(g_sqDtmfDetector[nDev].hLiDspPipe, result);
  }                  

  // connect pipes for the DTMF Detector
  for (nDev = 0; nDev < DEMO_NUMBER_OF_DEVS; nDev++) {
    
    // connect pipe, see DemoInitializeGlobalData()
    nTsLi  = g_sqDtmfDetector[nDev].nTs;
    nLi    = g_sqDtmfDetector[nDev].nLi;
    nDsp   = g_sqDtmfDetector[nDev].nDsp;

    if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
      // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots of the DSP highway.
      // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
      nTsDsp = (nTsLi * 4) + (nLi * 2) ;  // See comment above regarding the timeslot mapping for the DSP highway on this board
    } else
    if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
      // The Li0 timeslots on this card are the TS0, TS4, TS8,  ... , TS124 timeslots of the DSP highway.
      // The Li1 timeslots on this card are the TS1, TS5, TS9,  ... , TS125 timeslots of the DSP highway.    
      // The Li2 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the DSP highway.    
      // The Li3 timeslots on this card are the TS3, TS7, TS11, ... , TS127 timeslots of the DSP highway.    
      nTsDsp = (nTsLi * 4) + nLi ;  // See comment above regarding the timeslot mapping for the DSP highway on this board
    }


    memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqLiTimeSlotMask[nTsLi] = 0xFF;
    memset(sqDspTimeSlotMask, 0, coTsPerDspHwy * sizeof(OTX_UINT8));
    sqDspTimeSlotMask[nTsDsp] = 0xFF;

#if 0 //debug
    DemoLogf("RX[%.2d] Li=%.2d LiTs=%.2d DspTs=%.3d Dsp=%.1d \n", 
      nDev,
      g_sqDtmfDetector[nDev].nLi,       
      g_sqDtmfDetector[nDev].nTs,       
      nTsDsp,
      g_sqDtmfDetector[nDev].nDsp      
    );
#endif

    result = OtxDrvConnectPipe(g_sqDtmfDetector[nDev].hLiDspPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(g_sqDtmfDetector[nDev].hLiDspPipe, result);
    result = OtxDrvConnectPipe(g_sqDtmfDetector[nDev].hLiDspPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(g_sqDtmfDetector[nDev].hLiDspPipe, result);
  }


  // CONNECT LOGICAL DEVICES WITHIN THE DSP
  for (nDev = 0; nDev < DEMO_NUMBER_OF_DEVS; nDev++) {

    // Create the following Connections in receive direction:
    //  Li  ------> DtmfDetector
    result = OtxDrvConnectLogicalDevice(g_sqDtmfDetector[nDev].hLiDspPipe, -1, g_sqDtmfDetector[nDev].hHandle, -1); DemoCheckRc(g_sqDtmfDetector[nDev].hHandle, result);

  }

  // CONNECT LOGICAL DEVICES WITHIN THE DSP
  for (nDev = 0; nDev < DEMO_NUMBER_OF_DEVS; nDev++) {

    // Create the following Connections in receive direction:
    //  Li  ------> DualToneDetector
    result = OtxDrvConnectLogicalDevice(g_sqDtmfDetector[nDev].hLiDspPipe, -1, g_sqDualToneDetector[nDev].hHandle, -1); DemoCheckRc(g_sqDualToneDetector[nDev].hHandle, result);

  }

  // CONNECT LOGICAL DEVICES WITHIN THE DSP
  for (nDev = 0; nDev < DEMO_NUMBER_OF_DEVS; nDev++) {

    // Create the following Connections in receive direction:
    //  Li  ------> SilenceDetector
    result = OtxDrvConnectLogicalDevice(g_sqDtmfDetector[nDev].hLiDspPipe, -1, g_sqSilenceDetector[nDev].hHandle, -1); DemoCheckRc(g_sqSilenceDetector[nDev].hHandle, result);
  }

  // CONNECT LOGICAL DEVICES WITHIN THE DSP
  for (nDev = 0; nDev < DEMO_NUMBER_OF_DEVS; nDev++) {

    // Create the following Connections in receive direction:
    //  Li  ------> SingleToneDetector
    result = OtxDrvConnectLogicalDevice(g_sqDtmfDetector[nDev].hLiDspPipe, -1, g_sqSingleToneDetector[nDev].hHandle, -1); DemoCheckRc(g_sqSilenceDetector[nDev].hHandle, result);
  }

}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Initialize the DTMF Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {

    result = OtxToneDtmfDetectorInit(
      g_sqDtmfDetector[i].hHandle, 
      g_demoEnv.law, 
      OTX_TRUE); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);
  }


  // Initialize the Dual Tone Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {

    OTX_UINT32 nThreshold;
      
    result = OtxDrvGetAttributeValueUint32(g_sqDualToneDetector[i].hHandle, OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_TONE_ENERGY, &nThreshold); DemoCheckRc(g_sqDualToneDetector[i].hHandle, result);
    DemoLogf("Dual Tone Detector[%d]: OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
      
    result = OtxDrvGetAttributeValueUint32(g_sqDualToneDetector[i].hHandle, OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY, &nThreshold); DemoCheckRc(g_sqDualToneDetector[i].hHandle, result);
    DemoLogf("Dual Tone Detector[%d]: OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY = %d\n", i, nThreshold);
      
    result = OtxDrvGetAttributeValueUint32(g_sqDualToneDetector[i].hHandle, OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_TWIST, &nThreshold); DemoCheckRc(g_sqDualToneDetector[i].hHandle, result);
    DemoLogf("Dual Tone Detector[%d]: OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_TWIST        = %d\n", i, nThreshold);
      
    result = OtxDrvGetAttributeValueUint32(g_sqDualToneDetector[i].hHandle, OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_RELATIVE_3RD, &nThreshold); DemoCheckRc(g_sqDualToneDetector[i].hHandle, result);
    DemoLogf("Dual Tone Detector[%d]: OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_RELATIVE_3RD = %d\n", i, nThreshold);


    // Modifing default thresholds
    nThreshold = 600; // Min tone energy for first and second tones
    DemoLogf("Dual Tone Detector[%d]: Setting OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
    result = OtxDrvSetAttributeValueUint32(g_sqDualToneDetector[i].hHandle, OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_TONE_ENERGY, nThreshold); DemoCheckRc(g_sqDualToneDetector[i].hHandle, result);
      
    nThreshold = 200;  // Max energy for pause
    DemoLogf("Dual Tone Detector[%d]: Setting OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY = %d\n", i, nThreshold);
    result = OtxDrvSetAttributeValueUint32(g_sqDualToneDetector[i].hHandle, OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY, nThreshold); DemoCheckRc(g_sqDualToneDetector[i].hHandle, result);
      
    nThreshold = 300; // Max ratio HighestEnergy/2ndHighestEnergy*100 (225 = 2.25 times)
    // Note: We need a to accept a high ratio here since the frequencies for the
    // ring tone are very close to each other, which produces a large ratio between
    // the highest and the second highest energy level.
    DemoLogf("Dual Tone Detector[%d]: Setting OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_TWIST        = %d\n", i, nThreshold);
    result = OtxDrvSetAttributeValueUint32(g_sqDualToneDetector[i].hHandle, OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_TWIST, nThreshold); DemoCheckRc(g_sqDualToneDetector[i].hHandle, result);
      
    nThreshold = 30;  // Min ratio 2ndHighestEnergy/3rdHighestEnergy*100  (30 = 0.3 times)
    // Note: We need a to accept a low ratio here since the frequencies for the
    // dial tone are fairly low frequencies, which produces a relatively low ratio between
    // the second highest and the third highest energy level.
    DemoLogf("Dual Tone Detector[%d]: Setting OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_RELATIVE_3RD = %d\n", i, nThreshold);
    result = OtxDrvSetAttributeValueUint32(g_sqDualToneDetector[i].hHandle, OTX_ATTR_DUAL_TONE_DETECTOR_THRESHOLD_RELATIVE_3RD, nThreshold); DemoCheckRc(g_sqDualToneDetector[i].hHandle, result);

  }
    
    
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
      
    OTX_UINT32    sqFreq[4];       // An array of up to 6 frequencies to be detected.
    OTX_INT32     coFreq = 4;       // Number of elements in <p sqFreq>
      
#if 1 // Include this code to get a more status messages from the Dual Tone Detector
    if (i == 0) {
      result = OtxDrvSetAttributeValueUint32(
        g_sqDualToneDetector[i].hHandle, 
        OTX_ATTR_DUAL_TONE_DETECTOR_VERBOSE_STATUS, 
        0); 
      DemoCheckRc(g_sqDualToneDetector[i].hHandle, result);
    }
#endif

    /*
      
          .         FreqA  FreqB  FreqC  FreqD  FreqE  FreqF 
          FreqA    { -1,    1,      2,     4,     7,    11  },
          FreqB    {  1,   -1,      3,     5,     8,    12  },
          FreqC    {  2,    3,     -1,     6,     9,    13  },
          FreqD    {  4,    5,      6,    -1,    10,    14  },
          FreqE    {  7,    8,      9,    10,    -1,    15  },
          FreqF    { 11,   12,     13,    14,    15,    -1  }

      */

    // 350/440 is dial tone (dualtone index 1)
    // 440/480 is ring tone (dualtone index 3)
    // 480/620 is busy tone (dualtone index 6)

    sqFreq[0] = 350;  // Freq A in Hz
    sqFreq[1] = 440;  // Freq B in Hz       
    sqFreq[2] = 480;  // Freq C in Hz
    sqFreq[3] = 620;  // Freq D in Hz       
      
      
    // Initialize the Generic Dual Tone Detector. Use a couple of DTMF tone frequencies as an example
    result =  OtxToneDualDetectorInit(
      g_sqDualToneDetector[i].hHandle,           // Handle to the Logical device
      g_demoEnv.law,    // Data Format of the Monitored Data Stream
      sqFreq,           // An array of up to 6 frequencies to be detected.
      coFreq,           // Number of elements in <p sqFreq>
      OTX_TRUE,         // If OTX_TRUE, generate events and store the received digits
      OTX_TRUE,         // If OTX_TRUE the detector generates an event when a tone ends
      100               // Minimum duration (in milliseconds) required to validate a dual tone
      );
    DemoCheckRc(g_sqDualToneDetector[i].hHandle, result);
  }


  // Silence Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {

    // Initialize Silence Tone Detector
    result = OtxToneSilenceDetectorInit(
      g_sqSilenceDetector[i].hHandle, 
      g_demoEnv.law, 
      2, 
      -28); 
    DemoCheckRc(g_sqSilenceDetector[i].hHandle, result);

  }

  // Initialize the Single Tone Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {

    OTX_UINT32 nThreshold;
          
    result = OtxDrvGetAttributeValueUint32(g_sqSingleToneDetector[i].hHandle, OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY, &nThreshold); DemoCheckRc(g_sqSingleToneDetector[i].hHandle, result);
    DemoLogf("Single Tone Detector[%d]: OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
    
    result = OtxDrvGetAttributeValueUint32(g_sqSingleToneDetector[i].hHandle, OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_VARIANCE_ENERGY, &nThreshold); DemoCheckRc(g_sqSingleToneDetector[i].hHandle, result);
    DemoLogf("Single Tone Detector[%d]: OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_VARIANCE_ENERGY  = %d\n", i, nThreshold);
    
    result = OtxDrvGetAttributeValueUint32(g_sqSingleToneDetector[i].hHandle, OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY, &nThreshold); DemoCheckRc(g_sqSingleToneDetector[i].hHandle, result);
    DemoLogf("Single Tone Detector[%d]: OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY = %d\n", i, nThreshold);
    
    // Modifing default thresholds
    nThreshold = 1000;
    DemoLogf("Single Tone Detector[%d]: Setting OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
    result = OtxDrvSetAttributeValueUint32(g_sqSingleToneDetector[i].hHandle, OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY, nThreshold); DemoCheckRc(g_sqSingleToneDetector[i].hHandle, result);
    
    nThreshold = 200;
    DemoLogf("Single Tone Detector[%d]: Setting OTX_ATTR_SINGLE_VARIANCE_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
    result = OtxDrvSetAttributeValueUint32(g_sqSingleToneDetector[i].hHandle, OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_VARIANCE_ENERGY, nThreshold); DemoCheckRc(g_sqSingleToneDetector[i].hHandle, result);
    
    nThreshold = 200;
    DemoLogf("Single Tone Detector[%d]: Setting OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY = %d\n", i, nThreshold);
    result = OtxDrvSetAttributeValueUint32(g_sqSingleToneDetector[i].hHandle, OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY, nThreshold); DemoCheckRc(g_sqSingleToneDetector[i].hHandle, result);
          
  }

  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
  
    OTX_UINT32    sqFreq[2];       // An array of up to 6 frequencies to be detected.
    OTX_INT32     coFreq = 2;       // Number of elements in <p sqFreq>
          
#if 0 // Include this code to get a more status messages from the Single Tone Detector
    result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_VERBOSE_STATUS, 5); DemoCheckRc(g_hLogicDevs.hSingleToneDetector[i].hHandle, result);
#endif
          
    sqFreq[0] = 440;  // Freq A in Hz
    sqFreq[1] = 620;  // Freq B in Hz       
                   
    // Initialize the Generic Single Tone Detector. Use a couple of DTMF tone frequencies as an example
    result =  OtxToneSingleDetectorInit(
      g_sqSingleToneDetector[i].hHandle,           // Handle to the Logical device
      g_demoEnv.law,    // Data Format of the Monitored Data Stream
      sqFreq,           // An array of up to 6 frequencies to be detected.
      coFreq,           // Number of elements in <p sqFreq>
      OTX_TRUE,         // If OTX_TRUE, generate events and store the received digits
      OTX_TRUE,         // If OTX_TRUE the detector generates an event when a tone ends
      250                // Minimum duration (in milliseconds) required to validate a dual tone
      );
      DemoCheckRc(g_sqSingleToneDetector[i].hHandle, result);
          
  }

}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // DTMF Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
   
    result = OtxDrvEnable(g_sqDtmfDetector[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDtmfDetector[i].hHandle, result);
  }

  // Dual Tone Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {

    result = OtxDrvEnable(g_sqDualToneDetector[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDualToneDetector[i].hHandle, result);        
  }

  // Silence Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
   
    result = OtxDrvEnable(g_sqSilenceDetector[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqSilenceDetector[i].hHandle, result);    
  }

  // SingleTone Detector
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
   
    result = OtxDrvEnable(g_sqSingleToneDetector[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqSingleToneDetector[i].hHandle, result);    
  }

}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
  #endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {

    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ",
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min,
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }

}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*********************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;
}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DUAL_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dual Tone Detector
      if (pEventData->m_nCode == OTX_DUAL_TONE_DETECTOR_EC_TONE) {
        // 350/440 is dial tone (dualtone index 1)
        // 440/480 is ring tone (dualtone index 3)
        // 480/620 is busy tone (dualtone index 6)

        switch (pEventData->m_nParam) {
        case 1:
          DemoLogf("<DETECTION> DualDetector[%d] detected Dial Tone (%d) \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        case 3:
          DemoLogf("<DETECTION> DualDetector[%d] detected Ring Tone (%d) \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        case 6:
          DemoLogf("<DETECTION> DualDetector[%d] detected Busy Tone (%d) \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        default:
          DemoLogf("DualDetector[%d] detected unrecognized tone (%d) \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        }

      } else if (pEventData->m_nCode == OTX_DUAL_TONE_DETECTOR_EC_PAUSE) {
        DemoLogf("<DETECTION> DualDetector[%d] detected Pause\n", pEventData->m_nSourceId);
      } else if (pEventData->m_nCode == OTX_DUAL_TONE_DETECTOR_EC_VERBOSE_STATUS) {
        DemoLogf("DualDetector[%d] Verbose Status 0x%.8X (Code: %d, Value: %d)\n", pEventData->m_nSourceId, pEventData->m_nParam, pEventData->m_nParam & 0x0000FFFF, pEventData->m_nParam >> 16);
      } else {
        // Unexpected event from this device
        DemoLogf("DualDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("DualDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_SILENCE_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    
    if (OTX_SUCCESS(otxRc)) {
      // Event from silence Detector: OTX_SILENCE_DETECTOR_EC_SILENCE or OTX_SILENCE_DETECTOR_EC_NON_SILENCE
      DemoLogf("Silence Detector[%d]: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("Silence Detector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_SINGLE_DETECTOR:

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Single Tone Detector
      if (pEventData->m_nCode == OTX_SINGLE_TONE_DETECTOR_EC_TONE) {
        // sqFreq[0] = 440;  // Freq A in Hz
        // sqFreq[1] = 620;  // Freq B in Hz       

        switch (pEventData->m_nParam) {
        case 0:
          DemoLogf("SingleDetector[%d] detected 440Hz Tone (%d) \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        case 1:
          DemoLogf("SingleDetector[%d] detected 620Hz Tone (%d) \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        default:
          DemoLogf("SingleDetector[%d] detected unrecognized tone (%d) \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        }

      } else if (pEventData->m_nCode == OTX_SINGLE_TONE_DETECTOR_EC_PAUSE) {
        DemoLogf("SingleDetector[%d] detected Pause\n", pEventData->m_nSourceId);
      } else if (pEventData->m_nCode == OTX_SINGLE_TONE_DETECTOR_EC_VERBOSE_STATUS) {
        DemoLogf("SingleDetector[%d] Verbose Status 0x%.8X (Code: %d, Value: %d)\n", pEventData->m_nSourceId, pEventData->m_nParam, pEventData->m_nParam & 0x0000FFFF, pEventData->m_nParam >> 16);
      } else {
        // Unexpected event from this device
        DemoLogf("SingleDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("SingleDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

    default:
    // Unhandled event
    DemoLogf("Print Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }
}


/*******************************************************************************

  @func

  Detect DTMF digits

  @comm

*******************************************************************************/
void *DemoDtmfDetector(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_UINT32 nDev;
  int nIndex = 0;

  clock_t startTime = 0, stopTime;
  
  // validate parameters
  nDev = *((OTX_INT32 *) lpParameter);
  if (nDev < 0 || nDev > DEMO_NUMBER_OF_DEVS) {
    DemoLogf("DemoDtmfDetector: Thread was passed an invalid nDev(%d). Thread exiting...\n", nDev);
    return 0;
  }
        
  startTime = clock();

  g_sqDtmfDetector[nDev].bActive = OTX_TRUE;
    
    
  OtxDrvWaitForSingleEvent(g_sqDtmfDetector[nDev].hNotificationEvent, OTX_TIME_INFINITY);
    
  // Check for events from this device
  {
    OtxEventDataS eventData;
    OtxEventDataS *pEventData;
    char msgStr[128];
      
    do  {
      result = OtxDrvGetEventData(g_sqDtmfDetector[nDev].hEventQueue, &eventData); 
      OsaThreadSleep(1); // VERY IMPORTANT: give the driver some time, otherwise, CPU load is 100% 
        
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        pEventData = &eventData;
          
        switch (pEventData->m_eDeviceType) {
            
        case OTX_LDEVICE_TONE_DTMF_DETECTOR:
        result = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
          
        if (OTX_SUCCESS(result)) {
          // Event from Dtmf Detector
          if (pEventData->m_nCode == OTX_DTMF_DETECTOR_EC_DIGIT) {
            switch (pEventData->m_nParam) {
            case 0:  // 0  (941Hz/1336Hz)
            case 1:  // 1  (697Hz/1209Hz)
            case 2:  // 2  (697Hz/1336Hz)
            case 3:  // 3  (697Hz/1477Hz)
            case 4:  // 4  (770Hz/1209Hz)
            case 5:  // 5  (770Hz/1336Hz)
            case 6:  // 6  (770Hz/1477Hz)
            case 7:  // 7  (852Hz/1209Hz)
            case 8:  // 8  (852Hz/1336Hz)
            case 9:  // 9  (852Hz/1477Hz)
            DemoLogf("DtmfDetector[%d] detected DTMF digit %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
            break;

            case 10: // A  (697Hz/1633Hz)
            case 11: // B  (770Hz/1633Hz)
            case 12: // C  (852Hz/1633Hz)
            case 13: // D  (941Hz/1633Hz)
            DemoLogf("DtmfDetector[%d] detected DTMF key %C \n", pEventData->m_nSourceId, pEventData->m_nParam - 10 + 'A');
            break;

            case 14: // *  (941Hz/1209Hz)
            DemoLogf("DtmfDetector[%d] detected DTMF key '*'\n", pEventData->m_nSourceId);
            break;

            case 15: // #  (941Hz/1477Hz)
            DemoLogf("DtmfDetector[%d] detected DTMF key '#'\n", pEventData->m_nSourceId);
            break;

            default: 
            DemoLogf("DtmfDetector[%d] Unrecognized DTMF tone (%d)\n", pEventData->m_nSourceId, pEventData->m_nParam);
            break;
            }
          }

        } else {
          // Unexpected event from this device
          DemoLogf("DtmfDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
        }
        break;
            
        default:
          // Unhandled event
          DemoLogf("DemoDtmfDetector[%d]: Event from an unknown object 0x%X\n", nDev, pEventData->m_hDevice);
          OtxDrvDisable(pEventData->m_hDevice, OTX_TASK_SYNC, 5000L);
          break;
        }
      }
                
    } while(g_sqDtmfDetector[nDev].bActive);
    
  }
  
  stopTime = clock();
  
  DemoLogf("DemoDtmfDetector[%d] Thread Exiting (time=%2.1f seconds).\n", nDev, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
  
  return 0;
}



/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr      = NULL;
  envP->bExitOnError     = OTX_FALSE;
  envP->boardNo          = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType       = OTX_DEVICE_THOR_2_EXPRESSCARD;
  envP->coDsp            = 1;
  envP->coLi             = 2;
  envP->eT1E1Mode        = OTX_T1E1_MODE_E1;
  envP->eProgramModule   = OTX_DSP_SPM_OTXSPM2_INTMEM;
  envP->law              = OTX_DATA_FORMAT_COMPAND8_A_LAW;
  envP->nIteration       = 1;


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {          // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'I') {   // Test Iterations 
          envP->nIteration = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'P') {   // Dsp Program module
          if (*(argv[i]+2) == 'N') {          // OtxSpm2_Intmem.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2_INTMEM;
          } else {
            DemoLogf("Error: Invalid DSP Program Module command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'C') {
            envP->nBoardType = OTX_DEVICE_THOR_2_EXPRESSCARD;
            envP->coDsp      = 1;
            envP->coLi       = 2;
          } else if ( *(argv[i]+2) == 'R') {
            envP->nBoardType = OTX_DEVICE_THOR_4_EXPRESSCARD;
            envP->coDsp      = 1;
            envP->coLi       = 4;
          }else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {        // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') { // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {        // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') { // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}   



/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /T[C|R]        : Board Type used for the demo, where\n");
  DemoLogf("                  C = Use Thor-2-ExpressCard Board \n");
  DemoLogf("                  R = Use Thor-4-ExpressCard Board \n");
  DemoLogf(" /CA          : A-law\n");
  DemoLogf(" /CU          : u-law\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /P[N]        : DSP program to load, where\n");
  DemoLogf("                   N = OTX_DSP_SPM_OTXSPM2_INTMEM (default) \n");
  DemoLogf(" /I<count>    : Test Iteration \n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(0, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT result;
  static OTX_INT32      nCurrentLi = 1;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    
    break;


    case 'D':
    {
      OTX_INT32 i;
      int nDev = 0;
        
      printf("enter the device number: ");      
      scanf("%d",&nDev);      
      
      DemoLogf("\n nDev = %d", nDev);

      for (i = 0; i < 100; i++) {
          
        OTX_BOOL  bSilence = 0;      

        result = OtxToneSilenceDetectorStatus(g_sqSilenceDetector[nDev].hHandle, &bSilence); 
        DemoCheckRc(g_sqSilenceDetector[nDev].hHandle, result);

        // @parm OTX_TRUE if silence, OTX_FALSE if tone detected
        if (bSilence) {
          DemoLogf("s");
        } else {
          DemoLogf(".");
        }
        
        OsaThreadSleep(5);
      }
      
      DemoLogf("\n");
      OsaThreadSleep(1000);

      DemoLogf("\n");
      
    }
    break;


  case 'd':
    {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("get debug data from Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 0);
      }
    }
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;      // OTX Standard Return Code
  char commandChar;
  int i;
  
  OsaConsoleInit();

  DemoLogf("TecMultiToneDetect (Thor-2-ExpressCard), %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("DTMF, DualTone, SingleTone example program for the OTX Thor-2-ExpressCard.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  DemoInitializeGlobalData();

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  // Set Li0 as the clock source
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application, and connect the logical devices
  DemoLogf("Connecting logical devices...\n");
  DemoConnectDevices();
  DemoLogf("Connected logical devices\n");

  // Initialized the Connected Logical Devices
  DemoLogf("Logical devices are initializing...\n");
  DemoInitializeLogicalDevices();
  DemoLogf("Logical devices are initialized\n");

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Logical devices are enabling...\n");
  DemoEnableLogicalDevices();
  DemoLogf("Logical devices are enabled\n");

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);
  OsaThreadSleep(300); // To let DemoIdleFunction start up

  // start DTMF Detector threads
  DemoLogf("\nStarting DTMF Detector threads...\n");
  for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
    
    OsaThreadStartEx(DemoDtmfDetector,  (void *) &g_sqDtmfDetector[i].nDev);
    OsaThreadSleep(10);
  }


  DemoLogf("\nPress 'h' for help\n");
  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // abort threads
  DemoLogf("\nExiting...\n");
  {
    for (i = 0; i < DEMO_NUMBER_OF_DEVS; i++) {
      g_sqDtmfDetector[i].bActive = OTX_FALSE;
    }

    OsaThreadSleep(1000);   // give threads a chance to exit
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}

