/* $Workfile: T2pT1TR08DLAccess.c $       $Revision: 16 $      $Modtime: 1/02/07 6:31p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2pT1TR08DLAccess.c

    A simple example console application demonstrating the use of OTX Driver
    to access the DL bit and the T1 SLC96/TR-08 9-State signalling states 
    with the Thor-2-PCI, Thor-8-PCI, Thor-2-PCI-Plus, Thor-2-PCI-Express, Thor-8-PCI-Plus, or 
    the Thor-2-PCMCIA boards (STD, PLUS, CST, EX and PRO).

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    Channel 1 through 23 are used CAS information (referred to as CAS channels).
    Channel 24 is used the HDLC framing, so this timeslot is configured as
    a clear-channel.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"


#define DEMO_DEFAULT_LOG_FILE "T2pT1TR08DLAccess.log"
#define DEMO_DEFAULT_BOARD_NO 0
#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call


typedef struct DemoEnvT {
  OTX_BOOL        bExitOnError;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OTX_INT32       nBoardType;
  int             coLi;
  int             boardNo;
  OTX_BOOL        bChannelizedAbcdBitAccess; // If OTX_TRUE then report ABCD data per channel in the OTX_T1E1_EC_BITROB_SIGNALING event
  OTX_BOOL        bTr08BitAccessMode; // If OTX_TRUE then report ABCD data per channel in the OTX_T1E1_EC_TR08_STATE_CHANGE event

} DemoEnvT;

/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OtxT1RobBitDataS  g_sqSendRobBitData[8];
OTX_UINT8   g_sqHdlcFrame[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
OTX_INT32   g_nSleepTimeBetweenHdlcFrames = 1000; // in milliseconds
OTX_INT32   g_nSleepTimeBetweenBitrobEvents = 1000; // in milliseconds
OTX_INT32   g_cbHdlcFrame = 10;
OTX_INT32   g_coHdlcSenderFrames = 1;
OTX_BOOL    g_bSimulateBitRobSignalling = OTX_FALSE;
OTX_BOOL    g_bSimulateHdlcSignalling = OTX_FALSE;
#if 0 // for testing a full event queue condition
OTX_BOOL    g_bSimulateBusy = OTX_FALSE;
#endif
OTX_BOOL    g_bBitRobSignallingVerbose = OTX_TRUE;
OTX_BOOL    g_bHdlcSignallingVerbose = OTX_TRUE;
OTX_INT32   g_coBitRobSignallingEvents = 0;

struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2]; // Used for Thor-8-PCI and Thor-8-PCI-Plus (not for Thor-2-PCI)
  OTX_HANDLE hLi[8];
} g_hPhysDevs;


struct {
  OTX_HANDLE hLiIn[8];
  OTX_HANDLE hLiOut[8];
} g_hPipes;


struct {
  OTX_HANDLE hT1E1Cas[8];
  OTX_HANDLE hT1E1DlBitAccess[8];


  // The HDLC Receiver/Sender is connected to TS24 on the T1 span.
  OTX_HANDLE hHdlcSender[8];
  OTX_HANDLE hHdlcReceiver[8];
} g_hLogicDevs;


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;

/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(/*128*/16, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor-2-PCI board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-8-PCI and Thor-8-PCI-Plus

  // Connect to the Physical Thor-2 Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening OTX_DEVICE_THOR_PCMCIA board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Get handles to T1/E1 line interface devices 
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    // On these boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    // On the Thor-8-PCI and Thor-8-PCI-Plus boards the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    // Since the demo application assumes that Li0 is looped to Li1, enable Automatic CRC bit inversion for one of the Line Intefaces (not both)
    if (i == 0) {
      g_T1E1CfgData.t1CfgData.bAutoCrcBitInversion = OTX_FALSE;
    } else {
      g_T1E1CfgData.t1CfgData.bAutoCrcBitInversion = OTX_TRUE;
    }

    result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);

    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_T1); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    if (g_demoEnv.bChannelizedAbcdBitAccess) {
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_CHANNELIZED_ABCD_BIT_ACCESS, OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }  
      
    if (g_demoEnv.bTr08BitAccessMode) {
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_TR08_MODE, OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_CHANNELIZED_ABCD_BIT_ACCESS, OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[i], result);
	  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_AUTOMATIC_RESET, OTX_FALSE); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    }

  }

}



/*******************************************************************************
  @func
    Create the needed Logical Devices
  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;


  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create CAS Sender and Receiver logical devices hosted in the physical Line Interface devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hLi[i], OTX_LDEVICE_T1_ROB_BIT_ACCESS, g_hEventQueue, i, &g_hLogicDevs.hT1E1Cas[i]); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    // Create DL-Bit Access logical devices hosted in the physical Line Interface devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hLi[i], OTX_LDEVICE_T1_DL_BIT_ACCESS, g_hEventQueue, i, &g_hLogicDevs.hT1E1DlBitAccess[i]); DemoCheckRc(g_hPhysDevs.hLi[i], result);

  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hLi[i], OTX_LDEVICE_HDLC_SENDER, g_hEventQueue, i, &g_hLogicDevs.hHdlcSender[i]); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hLi[i], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 ts;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create an incoming and outgoing Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiIn[i]); DemoCheckRc(g_hPipes.hLiIn[i], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiOut[i]); DemoCheckRc(g_hPipes.hLiOut[i], result);

    // Connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
    ts = 24;
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiIn[i], result);
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[i], OTX_PIPE_INPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiOut[i], result);

    // Connect HDLC Sender and Receiver to the external T1/E1 Pipes
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiIn[i], -1, g_hLogicDevs.hHdlcReceiver[i], -1); DemoCheckRc(g_hPipes.hLiIn[i], result);
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hHdlcSender[i], -1, g_hPipes.hLiOut[i], -1); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);
  }
}


/*******************************************************************************
  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.
*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Initialize the Hdlc Sender and Receiver
    result = OtxHdlcSenderInit(
               g_hLogicDevs.hHdlcSender[i], // Handle to logical HDLC Filter Device
               OTX_HDLC_FFT_FLAGS,          // Fill character(s) to be send between HDLC frames
               2,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
               OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
               OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
               OTX_TRUE,                    // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
               OTX_HDLC_TR_NA);             // HDLC Transmission Rate for V.110 and X.30 modes
               DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);

    result = OtxHdlcReceiverInit(
               g_hLogicDevs.hHdlcReceiver[i], // Handle to logical HDLC Filter Device
               OTX_HDLC_FFT_ALL_ONES,          // Fill character(s) to be send between HDLC frames
               1,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
               OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
               OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
               OTX_TRUE,                    // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
               OTX_HDLC_TR_NA,              // HDLC Transmission Rate for V.110 and X.30 modes
               OTX_TRUE);                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
               DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);

  }
}


/*******************************************************************************

  @func
    Enable Configured Devices

*******************************************************************************/
void DemoEnableDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;

  // Enable Physical Devices
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // If board type is Thor-8-PCI or Thor-8-PCI-Plus then enable Quad E1T1 device
  if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS)) {
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

   // Enable Logical DL-Bit Devices
    result = OtxDrvEnable(g_hLogicDevs.hT1E1DlBitAccess[0],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hT1E1DlBitAccess[0], result);

  // Enable Logical T1/E1 Cas Devices
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hT1E1Cas[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hT1E1Cas[i], result);
  }

  // Enable Logical HDLC Devices 
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hHdlcSender[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);
    result = OtxDrvEnable(g_hLogicDevs.hHdlcReceiver[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);
  }
}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CAS_BR;     //try cc
  pT1CfgData->eTxLineCode                = OTX_T1_LC_AMI;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_AMI;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_F72;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x800000;  // 0x800000 means that channel 24 (1-count) is a clear channel. E.g. used for signalling

  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;
}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
#if 1
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
#else  // For testing a full event queue condition
    {
      OTX_UINT32 i;
      OTX_UINT32 coEvents;
      DemoLogf("Event signalled. Reading one event\n");
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) != OTX_S_OK) {
        DemoLogf("OtxDrvGetEventData() failed to return an event [1]\n");
        continue;
      } else {
        DemoPrintEvent(&eventData);
      }
      
      result = OtxDrvGetAttributeValueUint32(g_hEventQueue, OTX_ATTR_QUEUE_USAGE, &coEvents); DemoCheckRc(g_hEventQueue, result);
      for (i = 0; i < coEvents; i++) {
        DemoLogf("Reading event %u of %u\n", i+1, coEvents);

        while (g_bSimulateBusy) {
          Sleep(500);
        }
        result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);
        if (OTX_RESULT_CODE(result) != OTX_S_OK) {
          DemoLogf("OtxDrvGetEventData() failed to return an event [2]\n");
          break;
        } else {
          DemoPrintEvent(&eventData);
        }
      }
      DemoLogf("Sleeping until next event notification\n");
    }
#endif
  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
	OTX_RESULT otxRc;
	char msgStr[256];
	
	if (g_bBitRobSignallingVerbose) {
		// DemoLogf("%.8X%.8X ", pEventData->m_nTimeStamp.sUint32.nHigh, pEventData->m_nTimeStamp.sUint32.nLow);
	}
	
	switch (pEventData->m_eDeviceType) {
	case OTX_DEVICE_LI_T1E1:
		// Event from T1E1 physical device (Li)
		otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
		if (OTX_FAILURE(otxRc)) {
			DemoLogf("E1/T1[%d] No string for event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
		}
		if (pEventData->m_nParam) {
			DemoLogf("E1/T1[%d] Event: %s (0x%x)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
		} else {
			DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
		}
		
		break;
		
	//*******************************	
	case OTX_LDEVICE_T1_DL_BIT_ACCESS:
    switch (pEventData->m_nCode) {
    case OTX_T1E1_EC_DL_BITS_RECEIVED:
      {
        static OTX_UINT32 g_nLastReceivedDlBits[8];

        assert(pEventData->m_nSourceId < 8);

        // Print DL-bits if they have changed
        // Please note that this is not a typical DL-channel protocol implementation
        if (pEventData->m_nParam != g_nLastReceivedDlBits[pEventData->m_nSourceId]) {
          g_nLastReceivedDlBits[pEventData->m_nSourceId] = pEventData->m_nParam;
          DemoLogf("DL-Bit Device[%d] DL-Bits Received (0x%X)\n", pEventData->m_nSourceId, pEventData->m_nParam);
        }
      }
      break;
    case OTX_T1E1_EC_DL_BITS_SENT:
      {
        DemoLogf("DL-Bit Device[%d] DL-Bits Sent (0x%X)\n", pEventData->m_nSourceId, pEventData->m_nParam);

        // At this time it is ok to provide the driver with more DL-bits to be transmitted,
        // but we do not do that here in this simple demo application

      }
      break;
    default:
      DemoLogf("BitRob Device[%d] Unexpected Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      break;
    }
    break;

	//*******************
		
	case OTX_LDEVICE_T1_ROB_BIT_ACCESS:
		otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
		if ((pEventData->m_nCode == OTX_T1E1_EC_TR08_STATE_CHANGE) && g_demoEnv.bTr08BitAccessMode) {
			DemoLogf("TR08 Signalling Device[%d], TR08 State Change Received: Channel= %.2X State = %.2X\n", pEventData->m_nSourceId, pEventData->m_nParam >> 8, pEventData->m_nParam & 0xFF);
		}
		break;
		
		
		
	case OTX_LDEVICE_HDLC_RECEIVER:
		// Event from HDLC Receiver 
		if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
			OTX_UINT32 i;
			OTX_UINT8  buf[1024];
			OTX_UINT32 nFrameLen;
			OTX_UINT32 nFrameSeq;
			OtxHdlcFrameEndCodeE nFrameEndCode;
			OTX_DATETIME nTimeStamp;
			
			otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
			if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
				
				if (g_bHdlcSignallingVerbose) {
					DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);
					
					DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
					for (i = 0; i < nFrameLen && i < 1024; i++) {
						DemoLogf("%.2X ", buf[i]);
					}
					DemoLogf("\n");
				} else {
					if ((pEventData->m_nParam % 500) == 0) {
						DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);
					}
				}
				
			} else {
				DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
				DemoCheckRc(pEventData->m_hDevice, otxRc);
			}
			
			
		} else {
			otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
			if (OTX_SUCCESS(otxRc)) {
				DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
			} else {
				DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
			}
		}
		break;
		
	default:
	// Unhandled event
		DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
	break;
	}
	
	
}


/*******************************************************************************
  @func
    Extract a specific channel ABCD bit value from the OtxT1RobBitDataS struct.
*******************************************************************************/
char *DemoGetCasChannelValue(
  IN OTX_INT32 nCasCh,
  IN OtxT1RobBitDataS  *pRobBitData,
  OUT char *pszValue  // @parm ABCD bits (in that order) as a string
)
{
  pszValue[0] = '\0';
  
  if (nCasCh < 1 || nCasCh > 24) {
    DemoLogf("Error: CAS Channel out of range (%d)\n", nCasCh);
    return pszValue;
  }

  sprintf(pszValue + strlen(pszValue), "%d",  (pRobBitData->chA >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pRobBitData->chB >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pRobBitData->chC >> (nCasCh - 1)) & 0x01 ? 1 : 0);
  sprintf(pszValue + strlen(pszValue), "%d",  (pRobBitData->chD >> (nCasCh - 1)) & 0x01 ? 1 : 0);

  return pszValue;
}


/*******************************************************************************
  @func
    Prints the Rob Bit data for all 24 CAS channels
*******************************************************************************/
void DemoPrintAllCasChannelsData(
  OTX_INT32 nCasDev
)
{
  OTX_RESULT result;

  DemoLogf("Reading received bitrob data on LI#%d ... \n", nCasDev);

  if (g_demoEnv.bTr08BitAccessMode) {
    OTX_INT32 i;
    OTX_UINT8 nABCDBits;
   // OTX_UINT32 nABCDBits;



    for (i = 0; i < 24; i++) {
      result = OtxT1GetChannelizedRobBitData(g_hLogicDevs.hT1E1Cas[nCasDev], i+1, &nABCDBits); DemoCheckRc(g_hLogicDevs.hT1E1Cas[nCasDev], result);
      DemoLogf("   Channel %.2d: 0x%X", i+1, nABCDBits);
      if (i % 2) {
        DemoLogf("\n");
      }
    }

  } else {
    OTX_INT32 i;
    OtxT1RobBitDataS  rcvRobBitData;
    char szABCDbuf[4+1];  // +1 for the '\0'
    
    result = OtxT1GetRobBitData(g_hLogicDevs.hT1E1Cas[nCasDev], &rcvRobBitData); DemoCheckRc(g_hLogicDevs.hT1E1Cas[nCasDev], result);
    if (OTX_RESULT_CODE(result) != OTX_S_OK) {
      return;
    }
    DemoLogf("A = %.6X, B = %.6X, C = %.6X, D = %.6X\n", rcvRobBitData.chA, rcvRobBitData.chB, rcvRobBitData.chC, rcvRobBitData.chD);
    DemoLogf("\n");
    DemoLogf("            ABCD              ABCD\n");
    DemoLogf("            ====              ====\n");
    for (i = 0; i < 24; i++) {
      DemoLogf("Channel %.2d: ", i+1);
#if 0
      DemoLogf("%d",  (rcvRobBitData.chA >> i) & 0x01 ? 1 : 0);
      DemoLogf("%d",  (rcvRobBitData.chB >> i) & 0x01 ? 1 : 0);
      DemoLogf("%d",  (rcvRobBitData.chC >> i) & 0x01 ? 1 : 0);
      DemoLogf("%d",  (rcvRobBitData.chD >> i) & 0x01 ? 1 : 0);
#else
      DemoLogf("%s", DemoGetCasChannelValue(i+1, &rcvRobBitData, szABCDbuf));
#endif
      DemoLogf("  ");
      if (i % 2) {
        DemoLogf("\n");
      }
    }
    DemoLogf("\n");
  } 
}

/*******************************************************************************
  @func
    Send HDLC frames randomly
*******************************************************************************/
void *DemoSimulateHdlcSignalling(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_UINT8 sqHdlcFrame[128];
  OTX_DATETIME nTimeStamp;

  /* Seed the random-number generator with current time so that
   * the numbers will be different every time we run.
   */
  OtxDrvGetCurrentTime(&nTimeStamp);
  srand(nTimeStamp.sUint32.nLow);

  while (g_bSimulateHdlcSignalling) {
      OTX_INT32 nLi;
      OTX_INT32 cbFrameLen;
      OTX_INT32 i;
      
      nLi = rand() % g_demoEnv.coLi;

      cbFrameLen = (rand() % 30) + 3;

      for (i = 0; i < cbFrameLen; i++) {
        sqHdlcFrame[i] = (OTX_UINT8) i;
      }

      //DemoLogf("Sending HDLC frame from Li #%d (len=%d) ...", nLi, cbFrameLen);

      // Send HDLC Frame(s)
        result = OtxHdlcSendData(g_hLogicDevs.hHdlcSender[nLi], sqHdlcFrame, cbFrameLen, DEMO_TASK_REF, OTX_TIME_INFINITY); 
        if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
          // Driver IO Request buffer is full. Skip this frame.
        } else {
          DemoCheckRc(g_hLogicDevs.hHdlcSender[nLi], result);
        }


    OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
  }

  DemoLogf("DemoSimulateHdlcSignalling() thread is terminating.\n");
  return 0;

}




/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32 nCurrentCasChannel = 1;
  static OTX_INT32 nCurrentLi = 0;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-7 Select current Li. (currently Li=%d selected)\n", nCurrentLi);
    DemoLogf(" e/E Enable/Disable the DL-Bit Access Device on current Li\n");
    DemoLogf(" d   Transmit DL-Bit data on current Li\n");
    DemoLogf(" n   Increment current CAS channel (currently channel %d selected\n", nCurrentCasChannel);
    DemoLogf(" N   Decrement current CAS channel (currently channel %d selected\n", nCurrentCasChannel);
    DemoLogf(" r   Read the value of the received ABCD-bits on the other Li (all CAS channels)\n");
    DemoLogf(" R   Read the value of the received ABCD-bits on the current Li (all CAS channels)\n");
    DemoLogf(" l   Query Line Status of all Lis\n");
    DemoLogf(" t   Transmit HDLC frame on current Li (TS24)\n");
    DemoLogf(" u   Set time between each automatic bitrob event\n");
    DemoLogf(" v   Toggle verbose bitrob event reporting\n");
    DemoLogf(" V   Toggle verbose HDLC event reporting\n");
    DemoLogf(" j/J Start/Stop Automatic simulation of HDLC signalling\n");
    DemoLogf(" S   Set Sleep time between each HDLC frame\n");
    DemoLogf(" L   Set length of HDLC frame to be transmitted\n");
#if 0 // for testing a full event queue condition
    DemoLogf("\ng_bSimulateBusy=%d\n", g_bSimulateBusy);
#endif
    break;

  case 'v':
    {
      g_bBitRobSignallingVerbose = !g_bBitRobSignallingVerbose;
      DemoLogf("g_bBitRobSignallingVerbose is now %s.\n", g_bBitRobSignallingVerbose ? "TRUE": "FALSE");
    }
    break;

  case 'V':
    {
      g_bHdlcSignallingVerbose = !g_bHdlcSignallingVerbose;
      DemoLogf("g_bHdlcSignallingVerbose is now %s.\n", g_bHdlcSignallingVerbose ? "TRUE": "FALSE");
    }
    break;

    case 'r':
      {
        OTX_INT32 nLi = nCurrentLi + 1;
        
        if ((nCurrentLi % 2) == 0) {
          // Even Li
          nLi = nCurrentLi + 1;
        } else {
          // Odd Li
          nLi = nCurrentLi - 1;
        }
        DemoPrintAllCasChannelsData(nLi);
      }
      break;

    case 'R':
      {
        DemoPrintAllCasChannelsData(nCurrentLi);
      }
      break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 i;

      for (i = 0; i < g_demoEnv.coLi; i++) {
        DemoLogf("Line status of Li #%d is ", i);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[i], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;


  case 't':
    {
      static OTX_UINT8 nFrameCount = 0;
      OTX_INT32 i;
      OTX_INT32 busyCount = 0;

      if (g_cbHdlcFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
        DemoLogf("Error: Frame Length too long\n");
        break;
      }

      for (i = 0; i < g_cbHdlcFrame; i++) {
        g_sqHdlcFrame[i] = (OTX_UINT8) i;
      }
      g_sqHdlcFrame[0] = ++nFrameCount;

      DemoLogf("Sending %d HDLC frame(s) from Li #%d (len=%d, nFrameCount=0x%X) ...", g_coHdlcSenderFrames, nCurrentLi, g_cbHdlcFrame, nFrameCount);

      // Send HDLC Frame(s)
      for (i = 0; i < g_coHdlcSenderFrames; i++) {
        result = OtxHdlcSendData(g_hLogicDevs.hHdlcSender[nCurrentLi], g_sqHdlcFrame, g_cbHdlcFrame, DEMO_TASK_REF, OTX_TIME_INFINITY); 
        if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
          // Driver IO Request buffer is full. We need to retransmit this frame.
          i--;
          busyCount++;
        } else {
          DemoCheckRc(g_hLogicDevs.hHdlcSender[nCurrentLi], result);
          OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
        }
      }

      DemoLogf("Done (BusyCount=%d).\n", busyCount);
    }
    break;


  case 'L':
    {
      char strBuf[12];
      DemoLogf("Enter HDLC Frame Length to be transmitted (%d) ", g_cbHdlcFrame);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_cbHdlcFrame = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("HDLC Frame Length is now set to %d\n", g_cbHdlcFrame);
      } else {
        DemoLogf("HDLC Frame Length remains unchanged\n");
      }
    }
    break;

  case 'S':
    {
      char strBuf[12];
      DemoLogf("Enter Time to Sleep (in ms) between transmitted HDLC Frames (%d) ", g_nSleepTimeBetweenHdlcFrames);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_nSleepTimeBetweenHdlcFrames = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("Sleep time is now set to %d ms\n", g_nSleepTimeBetweenHdlcFrames);
      } else {
        DemoLogf("Sleep time remains unchanged\n");
      }
    }
    break;

  case '+':
    {
      char strBuf[12];

      DemoLogf("Set the number of consecutive frame to sent when 't' are pressed (%d)\n", g_coHdlcSenderFrames);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_coHdlcSenderFrames = (OTX_INT32) strtol(strBuf, NULL, 0);
        DemoLogf("Consecutive frame count is now %d\n", g_coHdlcSenderFrames);
      } else {
        DemoLogf("Consecutive frame count remains unchanged\n");
      }
    }
    break;


  case 'j':
    {
      DemoLogf("Starting Simulation of HDLC signalling\n");
      g_bSimulateHdlcSignalling = OTX_TRUE;
      OsaThreadStart(DemoSimulateHdlcSignalling);
      break;
    }

  case 'J':
    {
      DemoLogf("Aborting Simulation of HDLC signalling\n");
      g_bSimulateHdlcSignalling = OTX_FALSE;
      break;
    }

  case 'u':
    {
      char strBuf[12];
      DemoLogf("Enter Time to Sleep (in ms) between each Bitrob signalling event (%d) ", g_nSleepTimeBetweenBitrobEvents);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_nSleepTimeBetweenBitrobEvents = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("Sleep time is now set to %d ms\n", g_nSleepTimeBetweenBitrobEvents);
      } else {
        DemoLogf("Sleep time remains unchanged\n");
      }
    }
    break;



  case 'n':
    {
      // Select current nCurrentCasChannel
      nCurrentCasChannel++;
      if (nCurrentCasChannel > 24) {
        // Wrap 
        nCurrentCasChannel = 1;
      }
      
      DemoLogf("CAS Channel #%d Selected \n", nCurrentCasChannel);
      
    }
    break;
    
    
  case 'N':
    {
      // Select current nCurrentCasChannel
      nCurrentCasChannel--;
      if (nCurrentCasChannel < 1) {
        // Wrap
        nCurrentCasChannel = 24;
      }
      
      DemoLogf("CAS Channel #%d Selected \n", nCurrentCasChannel);
    }
    break;
    
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    {
      // Select current Li
      OTX_INT32 nLi = commandChar - '0';

      if (nLi > g_demoEnv.coLi) {
        DemoLogf("Error: LI#%d is not supported by this board\n", nLi);
      } else {
        nCurrentLi = nLi;
        DemoLogf("LI#%d Selected ....\n", nCurrentLi);
      }
    }
    break;
//********************
case 'e':
    {
      
      DemoLogf("Enable the T1/E1 DL-Bit Access Device on LI#%d \n", nCurrentLi);
      
      result = OtxDrvEnable(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi], result);
      
      // Provide the driver with initial DL-bits to be transmitted.
      // When these DL-bit have been transmitted the DL-Bit device will generate
      // a OTX_T1E1_EC_DL_BITS_SENT event. At that time we can provide more
      // DL-bits to the device to be transmitted.
      //result = OtxT1SetDlBitData(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi], 0x000000); DemoCheckRc(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi], result);
      
    }
    break;
    
  case 'E':
    {
      
      DemoLogf("Disable the T1/E1 DL-Bit Access Device on LI#%d \n", nCurrentLi);
      
      result = OtxDrvDisable(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi], result);
    }
    break;
    
  case 'd':
    {
      char szBuf[12]; 
      static OTX_UINT32 nDlBits = 0;
      
      DemoLogf("Enter DL-Bit data to transmit on LI#%d [0x%X]: ", nCurrentLi, nDlBits);
      OsaConsoleGetStr(szBuf);
      if (*szBuf) {
        nDlBits = strtoul(szBuf, NULL, 0);
      }
      
      // Provide the driver with initial DL-bits to be transmitted.
      // When these DL-bit have been transmitted the DL-Bit device will generate
      // a OTX_T1E1_EC_DL_BITS_SENT event. At that time we can provide more
      // DL-bits to the device to be transmitted.
      DemoLogf("Sending DL-Bit data: 0x%X\n", nDlBits);
      result = OtxT1SetDlBitData(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi], nDlBits); DemoCheckRc(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi], result);
      
    }
    break;

//********************
  case 'i':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'o':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;



  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************
  @func
    Parse the program's command line arguments and calls upon
    requested actions.

    There are some optional switches that can be used following
    the driver name on the command line. Switches are are
    separated by spaces, and can appear in arbitrary order.
    Switches consist of a hyphen and a letter:
*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->bExitOnError              = OTX_FALSE;
  envP->boardNo                   = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType                = OTX_DEVICE_THOR_PCI;
  envP->coLi                      = 2;
  envP->bChannelizedAbcdBitAccess = OTX_TRUE;
  envP->bTr08BitAccessMode        = OTX_TRUE;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    OsaStdStrUpr(argv[i]);
    if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
      if (*(argv[i]+1) == 'L') {   // Log file
        strcpy(envP->logFileName, argv[i] + 2);
      } else if (*(argv[i]+1) == 'B') {   // Board number
        envP->boardNo = strtoul(argv[i]+2, NULL, 0);
      } else if (*(argv[i]+1) == 'U') {   // Un-Channelized ABCD bit access
        envP->bChannelizedAbcdBitAccess = OTX_FALSE;
      } else if (*(argv[i]+1) == 'T') {   // Board Type
        if ( *(argv[i]+2) == '2') {
          envP->nBoardType = OTX_DEVICE_THOR_PCI;
          envP->coLi = 2;
        } else if ( *(argv[i]+2) == 'T') {
          envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;
          envP->coLi = 2;
        } else if ( *(argv[i]+2) == 'E') {
          envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_EXPRESS;
          envP->coLi = 2;
        } else if ( *(argv[i]+2) == '8') {
          envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
          envP->coLi = 8;
        } else if ( *(argv[i]+2) == 'P') {
          envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
          envP->coLi = 8;
        } else if ( *(argv[i]+2) == 'M') {
          envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
          envP->coLi = 2;
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
        
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
				exit(1);
      }
    } else {
      DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
      DemoHelpPanel();
      OsaConsoleExit();
			exit(1);
    }
  }

}


/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /T[2|T|E|8|P|M]  : Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI Board (default)\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                  E = Use Thor-2-PCI-Express Board\n");
  DemoLogf("                  8 = Use Thor-8-PCI Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                  M = Use Thor-2-PCMCIA Board\n");
  DemoLogf(" /U           : Use un-channelized ABCD bit access\n");
}


/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}

/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;
  OTX_INT32 i;

	OsaConsoleInit();

  DemoLogf("T2pT1RobBit, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Thor-2-PCI Example program for accessing T1 Rob-bit data and DL bit access\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  DemoLogf("ABCD Bit access is %s\n", g_demoEnv.bChannelizedAbcdBitAccess ? "channelized" : "non-channelized");

  for (i = 0; i < g_demoEnv.coLi; i++) {
    memset(&g_sqSendRobBitData[i], 0, sizeof(OtxT1RobBitDataS));
  }

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Set clock source for the board (important to set the clock source before configuring the E1/T1 devices)
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable all devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300);
  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/TM", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
